@@ -10,10 +10,10 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	10 April 2001
- * Last Edited:	16 August 2001
+ * Last Edited:	21 October 2003
  * 
  * The IMAP toolkit provided in this Distribution is
- * Copyright 2001 University of Washington.
+ * Copyright 1988-2003 University of Washington.
  * The full text of our legal notices is contained in the file called
  * CPYRIGHT, included with this Distribution.
  */
@@ -21,6 +21,10 @@
 #undef flock			/* name is used as a struct for fcntl */
 #undef fork			/* make damn sure that we don't use vfork!! */
 #include "nfstest.c"		/* get NFS tester */
+
+#ifndef NSIG			/* don't know if this can happen */
+#define NSIG 32			/* a common maximum */
+#endif
 
 /* Emulator for flock() call
  * Accepts: file descriptor
@@ -33,8 +37,8 @@ int flocksim (int fd,int op)
   char tmp[MAILTMPLEN];
   int logged = 0;
   struct flock fl;
-				/* lock applies to entire file */
-  fl.l_whence = fl.l_start = fl.l_len = 0;
+				/* lock zero bytes at byte 0 */
+  fl.l_whence = SEEK_SET; fl.l_start = fl.l_len = 0;
   fl.l_pid = getpid ();		/* shouldn't be necessary */
   switch (op & ~LOCK_NB) {	/* translate to fcntl() operation */
   case LOCK_EX:			/* exclusive */
@@ -74,7 +78,7 @@ int flocksim (int fd,int op)
 	sprintf (tmp,"Unexpected file locking failure: %s",strerror (errno));
 				/* give the user a warning of what happened */
 	MM_NOTIFY (NIL,tmp,WARN);
-	if (!logged++) syslog (LOG_ERR,tmp);
+	if (!logged++) syslog (LOG_ERR,"%s",tmp);
 	if (op & LOCK_NB) return -1;
 	sleep (5);		/* slow things down for loops */
       }
@@ -96,11 +100,68 @@ int flocksim (int fd,int op)
  * tested.  Fortunately, on BSD systems, OSF/1, and Linux, we can use the
  * flock() system call which doesn't have this bug.
  *
- *  Beware of systems such as AIX which offer flock() as a compatibility
+ *  Note that OSF/1, Linux, and some BSD systems have both broken fcntl()
+ * locking and the working flock() locking.
+ *
+ *  The program below can be used to demonstrate this problem.  Be sure to
+ * let it run long enough for all the sleep() calls to finish.
+ */
+
+#if 0
+#include <stdio.h>
+#include <unistd.h>
+#include <fcntl.h>
+#include <errno.h>
+#include <string.h>
+#include <sys/file.h>
+
+main ()
+{
+  struct flock fl;
+  int fd,fd2;
+  char *file = "a.a";
+  if ((fd = creat (file,0666)) < 0)
+    perror ("TEST FAILED: can't create test file"),_exit (errno);
+  close (fd);
+  if (fork ()) {		/* parent */
+    if ((fd = open (file,O_RDWR,0)) < 0) abort();
+				/* lock applies to entire file */
+    fl.l_whence = fl.l_start = fl.l_len = 0;
+    fl.l_pid = getpid ();	/* shouldn't be necessary */
+    fl.l_type = F_RDLCK;
+    if (fcntl (fd,F_SETLKW,&fl) == -1) abort ();
+    sleep (5);
+    if ((fd2 = open (file,O_RDWR,0)) < 0) abort ();
+    sleep (1);
+    puts ("parent test ready -- will hang here if locking works correctly");
+    close (fd2);
+    wait (0);
+    puts ("OS BUG: child terminated");
+    _exit (0);
+  }
+  else {			/* child */
+    sleep (2);
+    if ((fd = open (file,O_RDWR,0666)) < 0) abort ();
+    puts ("child test ready -- child will hang if no bug");
+				/* lock applies to entire file */
+    fl.l_whence = fl.l_start = fl.l_len = 0;
+    fl.l_pid = getpid ();	/* shouldn't be necessary */
+    fl.l_type = F_WRLCK;
+    if (fcntl (fd,F_SETLKW,&fl) == -1) abort ();
+    puts ("OS BUG: child got lock");
+  }
+}
+#endif
+
+/*  Beware of systems such as AIX which offer flock() as a compatibility
  * function that is just a jacket into fcntl() locking.  The program below
- * can be used to test if your system does this.  Be sure to let it run
- * long enough for all the sleep() calls to finish.  If the program hangs,
- * then you can dispense with the use of this module (you lucky fellow!).
+ * is a variant of the program above, only using flock().  It can be used
+ * to test to see if your system has real flock() or just a jacket into
+ * fcntl().
+ *
+ *  Be sure to let it run long enough for all the sleep() calls to finish.
+ * If the program hangs, then flock() works and you can dispense with the
+ * use of this module (you lucky person!).
  */
 
 #if 0
@@ -143,21 +204,23 @@ main ()
  *  On broken systems, we invoke an inferior fork to execute any driver
  * dispatches which are likely to tickle this bug; specifically, any
  * dispatch which may fiddle with a mailbox that is already selected.  As
- * of this writing, these are: delete, rename, status, copy, and append.
+ * of this writing, these are: delete, rename, status, scan, copy, and append.
  *
  *  Delete and rename are pretty marginal, yet there are certain clients
  * (e.g. Outlook Express) that really want to delete or rename the selected
- * mailbox.  The same is true of status, but there are people who don't
- * understand why status of the selected mailbox is bad news.
+ * mailbox.  The same is true of status, but there are people (such as the
+ * authors of Entourage) who don't understand why status of the selected
+ * mailbox is bad news.
  *
  *  However, in copy and append it is reasonable to do this to a selected
- * mailbox.
+ * mailbox.  Although scanning the selected mailbox isn't particularly
+ * sensible, it's hard to avoid due to wildcards.
  *
  *  It is still possible for an application to trigger the bug by doing
  * mail_open() on the same mailbox twice.  Don't do it.
  *
  *  Once the slave is invoked, the master only has to read events from the
- * slave's stdout (see below for these events) and translate these events
+ * slave's output (see below for these events) and translate these events
  * to the appropriate c-client callback.  When end of file occurs on the pipe,
  * the master reads the slave's exit status and uses that as the function
  * return.  The append master is slightly more complicated because it has to
@@ -171,38 +234,53 @@ main ()
  * by the slave since the master will generate these events for itself.
  *
  *  The other events cause the slave to write a newline-terminated string to
- * its stdout.  The first character of string indicates the event: S for
+ * its output.  The first character of string indicates the event: S for
  * mm_status(), N for mm_notify(), L for mm_log(), C for mm_critical(), X for
  * mm_nocritical(), D for mm_diskerror(), F for mm_fatal(), and "A" for append
  * argument callback.  Most of these events also carry data, which carried as
  * text space-delimited in the string.
  *
  *  Append argument callback requires the master to provide the slave with
- * data in the slave's stdin.  The first thing that the master provides is
+ * data in the slave's input.  The first thing that the master provides is
  * either a "+" (master has data for the slave) or a "-" (master has no data).
  * If the master has data, it will then send the flags, internal date, and
  * message text, each as <text octet count><SPACE><text>.
  */
+
+/*  It should be alright for lockslavep to be a global, since it will always
+ * be zero in the master (which is where threads would be).  The slave won't
+ * ever thread, since any driver which threads in its methods probably can't
+ * use fcntl() locking so won't have DR_LOCKING in its driver flags 
+ *
+ *  lockslavep can not be a static, since it's used by the dispatch macros.
+ */
 
 int lockslavep = 0;		/* non-zero means slave process for locking */
-
+static int lockproxycopy = 0;	/* non-zero means redo copy as proxy */
+FILE *slavein = NIL;		/* slave input */
+FILE *slaveout = NIL;		/* slave output */
+
+
 /* Common master
- * Accepts: append callback (append calls only, else NIL)
+ * Accepts: permitted stream
+ *	    append callback (append calls only, else NIL)
  *	    data for callback (append calls only, else NIL)
  * Returns: (master) T if slave succeeded, NIL if slave failed
  *	    (slave) NIL always, with lockslavep non-NIL
  */
 
-static long master (append_t af,void *data)
+static long master (MAILSTREAM *stream,append_t af,void *data)
 {
-  MAILSTREAM *stream;
+  MAILSTREAM *st;
   MAILSTATUS status;
   STRING *message;
   FILE *pi,*po;
+  blocknotify_t bn = (blocknotify_t) mail_parameters (NIL,GET_BLOCKNOTIFY,NIL);
   long ret = NIL;
-  long i;
+  unsigned long i,j;
   int c,pid,pipei[2],pipeo[2];
   char *s,*t,tmp[MAILTMPLEN];
+  lockproxycopy = NIL;		/* not doing a lock proxycopy */
 				/* make pipe from slave */
   if (pipe (pipei) < 0) mm_log ("Can't create input pipe",ERROR);
   else if (pipe (pipeo) < 0) {
@@ -215,62 +293,88 @@ static long master (append_t af,void *data)
     close (pipeo[0]); close (pipeo[1]);
   }
   else if (lockslavep = !pid) {	/* are we slave or master? */
-    alarm (0);			/* slave doesn't have alarms */
-    dup2 (pipeo[0],0);		/* parent's output in my input */
-    dup2 (pipei[1],1);		/* parent's input is my stdout */
+    alarm (0);			/* slave doesn't have alarms or signals */
+    for (c = 0; c < NSIG; c++) signal (c,SIG_DFL);
+    if (!(slavein = fdopen (pipeo[0],"r")) ||
+	!(slaveout = fdopen (pipei[1],"w")))
+      fatal ("Can't do slave pipe buffered I/O");
+    close (pipei[0]);		/* close parent's side of the pipes */
+    close (pipeo[1]);
   }
 
   else {			/* master process */
+    void *blockdata = (*bn) (BLOCK_SENSITIVE,NIL);
     close (pipei[1]);		/* close slave's side of the pipes */
     close (pipeo[0]);
     if (!(pi = fdopen (pipei[0],"r")) || !(po = fdopen (pipeo[1],"w")))
-      fatal ("Can't do pipe buffered I/O");
+      fatal ("Can't do master pipe buffered I/O");
 				/* do slave events until EOF */
 				/* read event */
     while (fgets (tmp,MAILTMPLEN,pi)) {
 				/* tie off event at end of line */
       if (s = strchr (tmp,'\n')) *s = '\0';
-      else fatal ("Execution process event string too long");
+      else {
+	s = "Execution process event string too long: ";
+	sprintf (t = (char *) fs_get (strlen (s) + strlen (tmp) + 1),"%s:%s",
+		 s,tmp);
+	fatal (t);
+      }
       switch (tmp[0]) {		/* analyze event */
       case 'A':			/* append callback */
 	if ((*af) (NIL,data,&s,&t,&message)) {
-	  putc ('+',po);	/* success */
-	  if (s) fprintf (po,"%ld %s",strlen (s),s);
-	  else fputs ("0 ",po);
-	  if (t) fprintf (po,"%ld %s",strlen (t),t);
-	  else fputs ("0 ",po);
-	  if (message) {
-	    fprintf (po,"%ld ",i = SIZE (message));
-	    if (i) do c = 0xff & SNX (message);
-	    while ((putc (c,po) != EOF) && --i);
-	    if (i) {
-	      sprintf (tmp,"Failed to write %lu bytes to execution process",i);
-	      fatal (tmp);
-	    }
+	  if (i = message ? SIZE (message) : 0) {
+	    if (!s) s = "";	/* default values */
+	    if (!t) t = "";
 	  }
-	  else fputs ("0 ",po);
+	  else s = t = "";	/* no flags or date if no message */
+	  errno = NIL;		/* reset last error */
+				/* build response */
+	  sprintf (tmp,"+%lu %s%lu %s%lu ",strlen (s),s,strlen (t),t,i);
+	  if (fputs (tmp,po) == EOF) {
+	    sprintf (tmp,"Failed to pipe command \"+%lu %s%lu %s%lu \": %s",
+		     strlen (s),s,strlen (t),t,i,strerror (errno));
+	    fatal (tmp);
+	  }
+				/* write message text */
+	  if (i) do if (putc (c = 0xff & SNX (message),po) == EOF) {
+	    sprintf (tmp,"Failed to pipe %lu bytes (of %lu), last=%u: %s",
+		     i,message->size,c,strerror (errno));
+	    fatal (tmp);
+	  } while (--i);
 	}
 	else putc ('-',po);	/* append error */
 	fflush (po);
 	break;
+      case '&':			/* slave wants a proxycopy? */
+	lockproxycopy = T;
+	break;
+
       case 'L':			/* mm_log() */
-	i = strtol (tmp+1,&s,10);
-	if (!s || (*s++ != ' ')) fatal ("Invalid log event arguments");
+	i = strtoul (tmp+1,&s,10);
+	if (!s || (*s++ != ' ')) {
+	  s = "Invalid log event arguments: ";
+	  sprintf (t = (char *) fs_get (strlen (s) + strlen (tmp) + 1),"%s:%s",
+		   s,tmp);
+	  fatal (t);
+	}
 	mm_log (s,i);
 	break;
       case 'N':			/* mm_notify() */
-	stream = (MAILSTREAM *) strtoul (tmp+1,&s,16);
+	st = (MAILSTREAM *) strtoul (tmp+1,&s,16);
 	if (s && (*s++ == ' ')) {
-	  i = strtol (s,&s,10);	/* get severity */
+	  i = strtoul (s,&s,10);/* get severity */
 	  if (s && (*s++ == ' ')) {
-	    mm_notify (stream,s,i);
+	    mm_notify ((st == stream) ? stream : NIL,s,i);
 	    break;
 	  }
 	}
-	fatal ("Invalid notify event arguments");
+	s = "Invalid notify event arguments: ";
+	sprintf (t = (char *) fs_get (strlen (s) + strlen (tmp) + 1),"%s:%s",
+		 s,tmp);
+	fatal (t);
 
       case 'S':			/* mm_status() */
-	stream = (MAILSTREAM *) strtoul (tmp+1,&s,16);
+	st = (MAILSTREAM *) strtoul (tmp+1,&s,16);
 	if (s && (*s++ == ' ')) {
 	  status.flags = strtoul (s,&s,10);
 	  if (s && (*s++ == ' ')) {
@@ -284,7 +388,7 @@ static long master (append_t af,void *data)
 		  if (s && (*s++ == ' ')) {
 		    status.uidvalidity = strtoul (s,&s,10);
 		    if (s && (*s++ == ' ')) {
-		      mm_status (stream,s,&status);
+		      mm_status ((st == stream) ? stream : NIL,s,&status);
 		      break;
 		    }
 		  }
@@ -293,39 +397,64 @@ static long master (append_t af,void *data)
 	    }
 	  }
 	}
-	fatal ("Invalid status event arguments");
+	s = "Invalid status event arguments: ";
+	sprintf (t = (char *) fs_get (strlen (s) + strlen (tmp) + 1),"%s:%s",
+		 s,tmp);
+	fatal (t);
       case 'C':			/* mm_critical() */
-	mm_critical ((MAILSTREAM *) strtoul (tmp+1,NIL,16));
-      case 'X':			/* mm_nocritical() */
-	mm_nocritical ((MAILSTREAM *) strtoul (tmp+1,NIL,16));
+	st = (MAILSTREAM *) strtoul (tmp+1,&s,16);
+	mm_critical ((st == stream) ? stream : NIL);
 	break;
+      case 'X':			/* mm_nocritical() */
+	st = (MAILSTREAM *) strtoul (tmp+1,&s,16);
+	mm_nocritical ((st == stream) ? stream : NIL);
+	break;
+
       case 'D':			/* mm_diskerror() */
-	stream = (MAILSTREAM *) strtoul (tmp+1,&s,16);
+	st = (MAILSTREAM *) strtoul (tmp+1,&s,16);
 	if (s && (*s++ == ' ')) {
-	  i = strtol (s,&s,10);
+	  i = strtoul (s,&s,10);
 	  if (s && (*s++ == ' ')) {
-	    fputc (mm_diskerror (stream,i,strtol (s,NIL,10)) ? '+' : '-',po);
-	    fflush (po);
+	    j = (long) strtoul (s,NIL,10);
+	    if (st == stream)	/* let's hope it's on usable stream */
+	      putc (mm_diskerror (stream,(long) i,j) ? '+' : '-',po);
+	    else if (j) {	/* serious diskerror on slave-created stream */
+	      mm_log ("Retrying disk write to avoid mailbox corruption!",WARN);
+	      sleep (5);	/* give some time for it to clear up */
+	      putc ('-',po);	/* don't abort */
+	    }
+	    else {		/* recoverable on slave-created stream */
+	      mm_log ("Error on disk write",ERROR);
+	      putc ('+',po);	/* so abort it */
+	    }
+	    fflush (po);	/* force it out either way */
 	    break;
 	  }
 	}
-	fatal ("Invalid diskerror event arguments");
+	s = "Invalid diskerror event arguments: ";
+	sprintf (t = (char *) fs_get (strlen (s) + strlen (tmp) + 1),"%s:%s",
+		 s,tmp);
+	fatal (t);
       case 'F':			/* mm_fatal() */
 	mm_fatal (tmp+1);
 	break;
       default:			/* random lossage */
-	fatal ("Unknown event from execution process");
+	s = "Unknown event from execution process: ";
+	sprintf (t = (char *) fs_get (strlen (s) + strlen (tmp) + 1),"%s:%s",
+		 s,tmp);
+	fatal (t);
       }
     }
-
     fclose (pi); fclose (po);	/* done with the pipes */
 				/* get slave status */
     grim_pid_reap_status (pid,NIL,&ret);
     if (ret & 0177) {		/* signal or stopped */
-      mm_log ("Execution process terminated abnormally",ERROR);
+      sprintf (tmp,"Execution process terminated abnormally (%lx)",ret);
+      mm_log (tmp,ERROR);
       ret = NIL;
     }
-    ret >>= 8;			/* return exit code */
+    else ret >>= 8;		/* return exit code */
+    (*bn) (BLOCK_NONSENSITIVE,blockdata);
   }
   return ret;			/* return status */
 }
@@ -342,7 +471,7 @@ static long master (append_t af,void *data)
 
 long safe_delete (DRIVER *dtb,MAILSTREAM *stream,char *mbx)
 {
-  long ret = master (NIL,NIL);
+  long ret = master (stream,NIL,NIL);
   if (lockslavep) exit ((*dtb->mbxdel) (stream,mbx));
   return ret;
 }
@@ -358,7 +487,7 @@ long safe_delete (DRIVER *dtb,MAILSTREAM *stream,char *mbx)
 
 long safe_rename (DRIVER *dtb,MAILSTREAM *stream,char *old,char *newname)
 {
-  long ret = master (NIL,NIL);
+  long ret = master (stream,NIL,NIL);
   if (lockslavep) exit ((*dtb->mbxren) (stream,old,newname));
   return ret;
 }
@@ -374,7 +503,7 @@ long safe_rename (DRIVER *dtb,MAILSTREAM *stream,char *old,char *newname)
 
 long safe_status (DRIVER *dtb,MAILSTREAM *stream,char *mbx,long flags)
 {
-  long ret = master (NIL,NIL);
+  long ret = master (stream,NIL,NIL);
   if (lockslavep) exit ((*dtb->status) (stream,mbx,flags));
   return ret;
 }
@@ -389,7 +518,7 @@ long safe_status (DRIVER *dtb,MAILSTREAM *stream,char *mbx,long flags)
 long safe_scan_contents (char *name,char *contents,unsigned long csiz,
 			 unsigned long fsiz)
 {
-  long ret = master (NIL,NIL);
+  long ret = master (NIL,NIL,NIL);
   if (lockslavep) exit (dummy_scan_contents (name,contents,csiz,fsiz));
   return ret;
 }
@@ -405,8 +534,16 @@ long safe_scan_contents (char *name,char *contents,unsigned long csiz,
 
 long safe_copy (DRIVER *dtb,MAILSTREAM *stream,char *seq,char *mbx,long flags)
 {
-  long ret = master (NIL,NIL);
-  if (lockslavep) exit ((*dtb->copy) (stream,seq,mbx,flags));
+  mailproxycopy_t pc =
+    (mailproxycopy_t) mail_parameters (stream,GET_MAILPROXYCOPY,NIL);
+  long ret = master (stream,NIL,NIL);
+  if (lockslavep) {
+				/* don't do proxycopy in slave */
+    if (pc) mail_parameters (stream,SET_MAILPROXYCOPY,(void *) slaveproxycopy);
+    exit ((*dtb->copy) (stream,seq,mbx,flags));
+  }
+				/* do any proxycopy in master */
+  if (lockproxycopy && pc) return (*pc) (stream,seq,mbx,flags);
   return ret;
 }
 
@@ -434,7 +571,7 @@ typedef struct append_data {
 long safe_append (DRIVER *dtb,MAILSTREAM *stream,char *mbx,append_t af,
 		  void *data)
 {
-  long ret = master (af,data);
+  long ret = master (stream,af,data);
   if (lockslavep) {
     APPENDDATA ad;
     ad.first = T;		/* initialize initial append package */
@@ -488,10 +625,10 @@ void slave_flags (MAILSTREAM *stream,unsigned long number)
 
 void slave_status (MAILSTREAM *stream,char *mailbox,MAILSTATUS *status)
 {
-  printf ("S%lx %ld %ld %ld %ld %ld %ld %s\n",
+  fprintf (slaveout,"S%lx %lu %lu %lu %lu %lu %lu %s\n",
 	  (unsigned long) stream,status->flags,status->messages,status->recent,
 	  status->unseen,status->uidnext,status->uidvalidity,mailbox);
-  fflush (stdout);
+  fflush (slaveout);
 }
 
 /* Notification event
@@ -502,8 +639,8 @@ void slave_status (MAILSTREAM *stream,char *mailbox,MAILSTATUS *status)
 
 void slave_notify (MAILSTREAM *stream,char *string,long errflg)
 {
-  printf ("N%lx %ld %s\n",(unsigned long) stream,errflg,string);
-  fflush (stdout);
+  fprintf (slaveout,"N%lx %lu %s\n",(unsigned long) stream,errflg,string);
+  fflush (slaveout);
 }
 
 
@@ -514,8 +651,8 @@ void slave_notify (MAILSTREAM *stream,char *string,long errflg)
 
 void slave_log (char *string,long errflg)
 {
-  printf ("L%ld %s\n",errflg,string);
-  fflush (stdout);
+  fprintf (slaveout,"L%lu %s\n",errflg,string);
+  fflush (slaveout);
 }
 
 
@@ -525,8 +662,8 @@ void slave_log (char *string,long errflg)
 
 void slave_critical (MAILSTREAM *stream)
 {
-  printf ("C%lx\n",(unsigned long) stream);
-  fflush (stdout);
+  fprintf (slaveout,"C%lx\n",(unsigned long) stream);
+  fflush (slaveout);
 }
 
 
@@ -536,8 +673,8 @@ void slave_critical (MAILSTREAM *stream)
 
 void slave_nocritical (MAILSTREAM *stream)
 {
-  printf ("X%lx\n",(unsigned long) stream);
-  fflush (stdout);
+  fprintf (slaveout,"X%lx\n",(unsigned long) stream);
+  fflush (slaveout);
 }
 
 /* Disk error found
@@ -550,10 +687,10 @@ void slave_nocritical (MAILSTREAM *stream)
 long slave_diskerror (MAILSTREAM *stream,long errcode,long serious)
 {
   int c;
-  printf ("D%lx %ld %ld\n",(unsigned long) stream,errcode,serious);
-  fflush (stdout);
-  if ((c = getchar ()) == '+') return LONGT;
-  if (c != '-') fatal ("Unknown master response for diskerror");
+  fprintf (slaveout,"D%lx %lu %lu\n",(unsigned long) stream,errcode,serious);
+  fflush (slaveout);
+  if ((c = getc (slavein)) == '+') return LONGT;
+  if (c != '-') slave_fatal ("Unknown master response for diskerror");
   return NIL;
 }
 
@@ -564,8 +701,10 @@ long slave_diskerror (MAILSTREAM *stream,long errcode,long serious)
 
 void slave_fatal (char *string)
 {
-  printf ("F%s\n",string);
-  exit (0);			/* die */
+  syslog (LOG_ALERT,"IMAP toolkit slave process crash: %.100s",string);
+  fprintf (slaveout,"F%s\n",string);
+  fflush (slaveout);
+  abort ();			/* die */
 }
 
 /* Append read buffer
@@ -579,7 +718,8 @@ static char *slave_append_read (unsigned long n,char *error)
 #if 0
   unsigned long i;
 #endif
-  char *t;
+  int c;
+  char *t,tmp[MAILTMPLEN];
   char *s = (char *) fs_get (n + 1);
   s[n] = '\0';
 #if 0
@@ -587,12 +727,14 @@ static char *slave_append_read (unsigned long n,char *error)
    * bug, since the problem only shows up if the application does fread()
    * on some other file
    */
-  for (t = s; n && ((i = fread (t,1,n,stdin)) || (errno == EINTR));
-       t += i,n -= i);
-  if (n) fatal (error);
+  for (t = s; n && ((i = fread (t,1,n,slavein)); t += i,n -= i);
 #else
-  for (t = s; n--; *t++ = getchar ());
+  for (t = s; n && ((c = getc (slavein)) != EOF); *t++ = c,--n);
 #endif
+  if (n) {
+    sprintf (tmp,"Error reading %s with %lu bytes remaining",error,n);
+    slave_fatal (tmp);
+  }
   return s;
 }
 
@@ -608,6 +750,7 @@ static char *slave_append_read (unsigned long n,char *error)
 long slave_append (MAILSTREAM *stream,void *data,char **flags,char **date,
 		   STRING **message)
 {
+  char tmp[MAILTMPLEN];
   unsigned long n;
   int c;
   APPENDDATA *ad = (APPENDDATA *) data;
@@ -616,22 +759,31 @@ long slave_append (MAILSTREAM *stream,void *data,char **flags,char **date,
   if (ad->date) fs_give ((void **) &ad->date);
   if (ad->msg) fs_give ((void **) &ad->msg);
   *flags = *date = NIL;		/* assume no flags or date */
-  puts ("A");			/* tell master we're doing append callback */
-  fflush (stdout);
-  switch (getchar ()) {		/* what did master say? */
+  fputs ("A\n",slaveout);	/* tell master we're doing append callback */
+  fflush (slaveout);
+  switch (c = getc (slavein)) {	/* what did master say? */
   case '+':			/* have message, get size of flags */
-    for (n = 0; isdigit (c = getchar ()); n *= 10, n += (c - '0'));
-    if (c != ' ') fatal ("Missing delimiter after flag size");
-    if (n) *flags = ad->flags = slave_append_read (n,"Error reading flags");
+    for (n = 0; isdigit (c = getc (slavein)); n *= 10, n += (c - '0'));
+    if (c != ' ') {
+      sprintf (tmp,"Missing delimiter after flag size %lu: %c",n,c);
+      slave_fatal (tmp);
+    }
+    if (n) *flags = ad->flags = slave_append_read (n,"flags");
 				/* get size of date */
-    for (n = 0; isdigit (c = getchar ()); n *= 10, n += (c - '0'));
-    if (c != ' ') fatal ("Missing delimiter after date size");
-    if (n) *date = ad->date = slave_append_read (n,"Error reading date");
+    for (n = 0; isdigit (c = getc (slavein)); n *= 10, n += (c - '0'));
+    if (c != ' ') {
+      sprintf (tmp,"Missing delimiter after date size %lu: %c",n,c);
+      slave_fatal (tmp);
+    }
+    if (n) *date = ad->date = slave_append_read (n,"date");
 				/* get size of message */
-    for (n = 0; isdigit (c = getchar ()); n *= 10, n += (c - '0'));
-    if (c != ' ') fatal ("Missing delimiter after message size");
+    for (n = 0; isdigit (c = getc (slavein)); n *= 10, n += (c - '0'));
+    if (c != ' ') {
+      sprintf (tmp,"Missing delimiter after message size %lu: %c",n,c);
+      slave_fatal (tmp);
+    }
     if (n) {			/* make buffer for message */
-      ad->msg = slave_append_read (n,"Error reading message");
+      ad->msg = slave_append_read (n,"message");
 				/* initialize stringstruct */
       INIT (&ad->message,mail_string,(void *) ad->msg,n);
       ad->first = NIL;		/* no longer first message */
@@ -641,8 +793,26 @@ long slave_append (MAILSTREAM *stream,void *data,char **flags,char **date,
     return LONGT;
   case '-':			/* error */
     *message = NIL;		/* set stop */
-    return NIL;			/* return failure */
+    break;
+  default:			/* unknown event */
+    sprintf (tmp,"Unknown master response for append: %c",c);
+    slave_fatal (tmp);
   }
-  fatal ("Unknown master response for append");
-  return NIL;
+  return NIL;			/* return failure */
+}
+
+/* Proxy copy across mailbox formats
+ * Accepts: mail stream
+ *	    sequence to copy on this stream
+ *	    destination mailbox
+ *	    option flags
+ * Returns: T if success, else NIL
+ */
+
+long slaveproxycopy (MAILSTREAM *stream,char *sequence,char *mailbox,
+		     long options)
+{
+  fputs ("&\n",slaveout);	/* redo copy as append */
+  fflush (slaveout);
+  return NIL;			/* failure for now */
 }
