@@ -1,5 +1,5 @@
 /*----------------------------------------------------------------------
-  $Id: headers.h,v 4.34 1999/02/09 22:52:50 hubert Exp $
+  $Id: headers.h 13927 2004-12-21 23:56:53Z hubert $
 
             T H E    P I N E    M A I L   S Y S T E M
 
@@ -20,7 +20,7 @@
    permission of the University of Washington.
 
    Pine, Pico, and Pilot software and its included text are Copyright
-   1989-1998 by the University of Washington.
+   1989-2000 by the University of Washington.
 
    The full text of our legal notices is contained in the file called
    CPYRIGHT, included with this distribution.
@@ -63,6 +63,22 @@ is os-xxx.h. (Don't edit osdep.h; edit os-xxx.h instead.)
 
 #include "os.h"
 
+#if !defined(DOS) && !defined(OS2) && !defined(LEAVEOUTFIFO)
+    /*
+     * If LEAVEOUTFIFO is set in os.h, then we leave it out.
+     * If it isn't set, we still might leave it out. We'll decide
+     * based on whether or not O_NONBLOCK is defined or not.
+     * It's just a guess. Safer would be to change the polarity of the
+     * test and define something like INCLUDEFIFO instead of LEAVEOUTFIFO
+     * and only define it where we know. Since we don't really know
+     * we'd rather run the risk of being wrong and finding out that
+     * way instead of just having people not know about it.
+     */
+#if !defined(O_NONBLOCK)
+#define LEAVEOUTFIFO 1
+#endif
+#endif
+
 #include "../c-client/rfc822.h"
 #include "../c-client/misc.h"
 
@@ -70,23 +86,18 @@ is os-xxx.h. (Don't edit osdep.h; edit os-xxx.h instead.)
 
 #include <lber.h>
 #include <ldap.h>
-/*
- * If LDAP_OPT_ON is defined in ldap.h, then this is the newer version of the
- * API (as opposed to the RFC1823 version). The newer version at the time
- * this was written was only described in draft-ietf-asid-ldap-c-api-00.txt.
- * Hmm. Seems mozilla.org defined LDAP_OPT_ON even in 1.1, so we'll require
- * that and LDAP_VERSION3.
- */
+
 #ifndef LDAPAPI
-#if	(LDAP_VERSION > 2)
-#define	LDAPAPI	20
-#else
-#if	defined(LDAP_OPT_ON)
-#define	LDAPAPI 11
-#else
-#define	LDAPAPI 10
-#endif
+#if defined(LDAP_API_VERSION)		/* draft-ietf-ldapext-ldap-c-api-04 */
+#define LDAPAPI LDAP_API_VERSION
+#elif defined(LDAP_OPT_SIZELIMIT)
+#define LDAPAPI 15			/* Netscape SDK */
+#elif defined(LDAP_BEGIN_DECL)
+#define LDAPAPI 11			/* OpenLDAP 1.x */
+#else					/* older version */
+#define LDAPAPI 10			/* Umich */
 #endif
+
 #ifndef LDAP_OPT_ON
 #define LDAP_OPT_ON ((void *)1)
 #endif
@@ -102,14 +113,13 @@ is os-xxx.h. (Don't edit osdep.h; edit os-xxx.h instead.)
 #ifndef LDAP_OPT_PROTOCOL_VERSION
 #define LDAP_OPT_PROTOCOL_VERSION 1136
 #endif
-#endif
 
 #ifndef LDAP_MSG_ONE
 #define LDAP_MSG_ONE (0x00)
 #define LDAP_MSG_ALL (0x01)
 #define LDAP_MSG_RECEIVED (0x02)
 #endif
-
+#endif
 #endif  /* ENABLE_LDAP */
 
 #include "helptext.h"
@@ -128,6 +138,7 @@ extern struct pine *ps_global;
 
 extern char	   *pine_version;	/* pointer to version string	     */
 
+#define SIZEOF_20KBUF (20480)
 extern char         tmp_20k_buf[];
 
 #ifdef DEBUG
